
import os
import copy
import cv2
import numpy as np
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import csv
from sklearn.cluster import OPTICS, DBSCAN, AgglomerativeClustering
import sklearn.metrics

from modules.file_utils import FileUtils
from modules.ply_reader import PlyReader
from modules.hough_transform import HoughTransform
from modules.helper_scripts import lines_to_features, features_to_lines, hex_to_rgb, generate_colours, fill_polygon
from modules.lines_utils import *


def plot(img, features, classes, unnormalize):

    for idx in range(len(unnormalize)):
        features[:, idx] = unnormalize[idx](features[:, idx])

    # for line in lines:
    #     x1 = line.start_point.x
    #     y1 = line.start_point.y
    #     x2 = line.end_point.x
    #     y2 = line.end_point.y
    #     cv2.line(img, (x1, y1), (x2, y2), (255, 0, 0), 3)

    colors = generate_colours(classes.max() + 1)
    for idx in range(features.shape[0]):
        x = features[idx, 0]
        y = features[idx, 1]

        if classes[idx] == -1:
            color = (0, 0, 0)
        else:
            color = hex_to_rgb(colors[classes[idx]])

        cv2.circle(img, (int(x), int(y)), radius=3, color=color, thickness=-1)

    # Show result
    cv2.imshow("Result Image", img)
    cv2.waitKey()  # cv2.imshow won't work without this :)
    cv2.destroyAllWindows()


def plot_poly(img, poly, show=True):
    img = copy.copy(img)
    for idx, point in enumerate(poly):
        p1 = [int(i) for i in point.as_list]
        if idx + 1 == len(poly):
            p2 = [int(i) for i in poly[0].as_list]
        else:
            p2 = [int(i) for i in poly[idx+1].as_list]
        cv2.line(img, p1, p2, (255, 0, 0), 3)
    if show:
        cv2.imshow("Poly image", img)
        cv2.waitKey(0)  # cv2.imshow won't work without this :)
        cv2.destroyAllWindows()
    return img


def plot_from_midpoint(args, img, centers):
    fig, ax = plt.subplots(1,1)
    canvas = fig.canvas
    ax.axis('off')

    ax.imshow(img)
    length = 500

    if args.is_probabalistic_hough:
        ax.scatter(centers[:,0], centers[:,1], s=5, c='r')
        for center in centers:
            starty = center[1] - length * np.sin(center[2])
            startx = center[0] - length * np.cos(center[2])
            endy = center[1] + length * np.sin(center[2])
            endx = center[0] + length * np.cos(center[2])
            ax.plot([startx, endx], [starty, endy], c='g')
    else:
        for center in centers:
            a = np.cos(center[1])
            b = np.sin(center[1])
            x0 = a * center[0]
            y0 = b * center[0]
            x1 = int(x0 + length*(-b))
            y1 = int(y0 + length*(a))
            x2 = int(x0 - length*(-b))
            y2 = int(y0 - length*(a))
            ax.plot([x1, x2], [y1, y2], c='g')
    plt.show(block=False)
    a = 0

def line_approximation(args, data):
    if args.line_approx_method == 'mean':
        return np.mean(data, axis=0)
    if args.line_approx_method == 'median':
        return np.median(data, axis=0)
    if args.line_approx_method == 'medoid':
        pairwise_distances = sklearn.metrics.pairwise_distances(X=data)
        medoid_idx = np.argmin(pairwise_distances.sum(axis=0))
        return data[medoid_idx]


def cluster(args, data):

    if args.clustering_algo == 'dbscan':
        clusterer = DBSCAN(eps=args.dbscan_eps,
                           min_samples=args.clustering_minSamples,
                           metric=args.clustering_metric,
                           metric_params=None,
                           algorithm='auto',
                           leaf_size=30,
                           p=None,
                           n_jobs=args.clustering_parallelJobCount)
        clusterer.fit(data)
        labels = clusterer.labels_


        # cluster center search
        centers = np.zeros((labels.max()+1, data.shape[1]))
        for idx_label in range(labels.max()+1):
            label_data = data[labels==idx_label]
            centers[idx_label] = line_approximation(args, label_data[:])

    elif args.clustering_algo == 'optics':
        clusterer = OPTICS(min_samples=args.clustering_minSamples,
                           max_eps=np.inf,
                           metric=args.clustering_metric,  # default minkowski
                           p=2,
                           cluster_method=args.optics_clusterMethod,
                           xi=0.05,
                           predecessor_correction=True,
                           min_cluster_size=args.clustering_minClusterSize,
                           algorithm='auto',
                           leaf_size=30,
                           n_jobs=args.clustering_parallelJobCount)
        clusterer.fit(data)
        labels = clusterer.labels_
        # cluster center search
        centers = np.zeros((labels.max() + 1, data.shape[1]))
        for idx_label in range(labels.max() + 1):
            label_data = data[labels == idx_label]
            centers[idx_label] = line_approximation(args, label_data[:])

    elif args.clustering_algo == 'agglomerative':
        clusterer = AgglomerativeClustering(n_clusters=None,
                                            distance_threshold=args.agglomerative_distThreshold,
                                            affinity=args.agglomerative_affinity)
        clusterer.fit(data)
        labels = clusterer.labels_
        # cluster center search
        centers = np.zeros((labels.max() + 1, data.shape[1]))
        for idx_label in range(labels.max() + 1):
            label_data = data[labels == idx_label]
            centers[idx_label] = line_approximation(args, label_data[:])



    return centers, labels


def lines_to_mask(args, map_image, lines):
    if map_image.ndim == 2:
        mask = np.zeros((map_image.shape), dtype=float)
    else:
        mask = np.zeros((map_image.shape[:2]), dtype=float)

    for line in lines:
        angle = line.angle
        p0 = (line.start_point.x + args.metric_line_width * np.cos(angle + np.pi/2), line.start_point.y + args.metric_line_width * np.sin(angle + np.pi/2))
        p1 = (line.start_point.x - args.metric_line_width * np.cos(angle + np.pi/2), line.start_point.y - args.metric_line_width * np.sin(angle + np.pi/2))
        p2 = (line.end_point.x - args.metric_line_width * np.cos(angle + np.pi/2), line.end_point.y - args.metric_line_width * np.sin(angle + np.pi/2))
        p3 = (line.end_point.x + args.metric_line_width * np.cos(angle + np.pi/2), line.end_point.y + args.metric_line_width * np.sin(angle + np.pi/2))
        coords = [p0, p1, p2, p3, p0]
        fill_polygon(coords, mask)
    return mask



def loss(gt_img, pred_img):
    return np.mean(np.abs(pred_img-gt_img))


def image_to_vertices(map_image):
    vertices = []
    for x in range(map_image.shape[0]):
        for y in range(map_image.shape[1]):
            if map_image[x, y, 0] != 255:
                vertices.append(
                    (y / map_image.shape[1], x / map_image.shape[0])
                )
    return np.array(vertices)


def process_ply(filepath_ply):
    output_dict = {
        'accessible_polygon': [],
        'vertices': []
    }
    try:
        args = FileUtils.load_json_as_object('./args.json')

        files_in = [filepath_ply]
        scores = {}
        for file_idx, file in enumerate(files_in):
            # if file_idx != len(files_in)-1:
            #     continue
            scores[file] = {}
            file_path_in = os.path.join(args.path_data_in, file)
            is_gt = False

            map_points_3d = PlyReader.read(file_path_in) # read input data

            map_image = PlyReader.points_3D_to_image(map_points_3d, show=False) # convert 3d points to image
            map_image = HoughTransform.detect_blobs(map_image, args, show=False)
            map_image = PlyReader.purge_points(map_image)
            map_image = HoughTransform.detect_blobs(map_image, args, show=False)
            map_image_copy = copy.copy(map_image)  # store copy image for final rotations of image
            # map_image = HoughTransform.rotate_image(map_image, show=False)

            lines = HoughTransform.detect_lines(map_image, args)  # detect hough lines
            img = HoughTransform.plot(map_image, lines, show=False)
            filename = file.replace('.ply', '_hough_lines.jpg')
            #cv2.imwrite(filename, img)

            filename = file.replace('.ply', '_map_image.jpg')
            #cv2.imwrite(filename, map_image)

            features, unnormalize = lines_to_features(args, lines)

            fuse_tries = 20
            while fuse_tries > 0:
                fuse_tries -= 1

                centers, labels = cluster(args, features)
                lines_filtered, lines_pixels = features_to_lines(args, map_image, centers, unnormalize)

                # HoughTransform.plot(map_image, lines_filtered, show=True)

                # filter lines that are in middle of other lines
                # filter close line that are parallel to other lines
                not_done = True
                while not_done:
                    is_edit_lines = False
                    idx_high_low = np.argsort(lines_pixels)[::-1]
                    for idx in idx_high_low:
                        for idx_second in range(len(lines_filtered)):
                            if lines_filtered[idx_second] == lines_filtered[idx]:
                                continue

                            intersection = lines_filtered[idx].intersects_line(lines_filtered[idx_second])
                            # check if intersenction in line middle
                            if intersection:
                                distance_midpoint = lines_filtered[idx].midpoint.distance(intersection)
                                distance_start = lines_filtered[idx].start_point.distance(intersection)
                                distance_end = lines_filtered[idx].end_point.distance(intersection)
                                if distance_midpoint < distance_start and distance_midpoint < distance_end:
                                    lines_filtered.pop(idx_second)
                                    lines_pixels.pop(idx_second)
                                    is_edit_lines = True
                                    break

                            # check how close lines are
                            delta_threshold = 10
                            start_point_delta_x = abs(lines_filtered[idx].start_point.x * np.cos(lines_filtered[idx].angle) -\
                                                  lines_filtered[idx_second].start_point.x * np.cos(lines_filtered[idx_second].angle))
                            start_point_delta_y = abs(lines_filtered[idx].start_point.y * np.sin(lines_filtered[idx].angle) - \
                                              lines_filtered[idx_second].start_point.y * np.sin(lines_filtered[idx_second].angle))
                            end_point_delta_x = abs(lines_filtered[idx].end_point.x * np.cos(lines_filtered[idx].angle) - \
                                                  lines_filtered[idx_second].end_point.x * np.cos(lines_filtered[idx_second].angle))
                            end_point_delta_y = abs(lines_filtered[idx].end_point.y * np.sin(lines_filtered[idx].angle) - \
                                                  lines_filtered[idx_second].end_point.y * np.sin(lines_filtered[idx_second].angle))
                            if start_point_delta_x < delta_threshold and start_point_delta_y < delta_threshold and \
                                    end_point_delta_x < delta_threshold and end_point_delta_y < delta_threshold:
                                lines_filtered.pop(idx_second)
                                lines_pixels.pop(idx_second)
                                is_edit_lines = True
                                break

                            if is_edit_lines:
                                break
                        if is_edit_lines:
                            break
                    if not is_edit_lines:
                        not_done = False

                # Close polygon
                lines_poly_search = copy.copy(lines_filtered)
                first_line = lines_poly_search.pop(np.argsort(lines_pixels)[::-1][0])
                # HoughTransform.plot(map_image, [first_line], show=True)
                polygon_points = [first_line.start_point, first_line.end_point]
                got_neighbour = [False, False]
                while not all(got_neighbour):
                    # plot_poly(map_image, polygon_points)
                    if not lines_poly_search:
                        break

                    best_poly_idx = None
                    best_dist = np.inf
                    best_neighbour_line_idx = None
                    neighbour_is_start = None
                    for idx_point in range(len(got_neighbour)):
                        if got_neighbour[idx_point]:
                            continue

                        if idx_point == 0:
                            other_point = polygon_points[-1]
                        else:
                            other_point = polygon_points[0]

                        current_point = polygon_points[idx_point]
                        for idx in range(len(lines_poly_search)):
                            start_dist = current_point.distance(lines_poly_search[idx].start_point)
                            if start_dist < best_dist:
                                best_poly_idx = idx_point
                                best_dist = start_dist
                                best_neighbour_line_idx = idx
                                neighbour_is_start = True

                            end_dist = current_point.distance(lines_poly_search[idx].end_point)
                            if end_dist < best_dist:
                                best_poly_idx = idx_point
                                best_dist = start_dist
                                best_neighbour_line_idx = idx
                                neighbour_is_start = False

                    distance_to_other_end = current_point.distance(other_point)
                    if distance_to_other_end < best_dist:
                        for idx_n in range(len(got_neighbour)):
                            got_neighbour[idx_n] = True
                    else:
                        if best_poly_idx == 0:
                            neighbour_line = lines_poly_search.pop(best_neighbour_line_idx)
                            new_points = [neighbour_line.start_point, neighbour_line.end_point]
                            if neighbour_is_start:
                                new_points = new_points[::-1]
                            polygon_points = new_points + polygon_points
                            got_neighbour[best_poly_idx] = True
                            got_neighbour = [False, True] + got_neighbour
                        else:
                            neighbour_line = lines_poly_search.pop(best_neighbour_line_idx)
                            new_points = [neighbour_line.end_point, neighbour_line.start_point]
                            if neighbour_is_start:
                                new_points = new_points[::-1]
                            polygon_points.extend(new_points)
                            got_neighbour[best_poly_idx] = True
                            got_neighbour.extend([True, False])

                # filter polygons

                # median distance between points
                dists = []
                for idx in range(1, len(polygon_points)):
                    dists.append(polygon_points[idx].distance(polygon_points[idx-1]))
                dist_median = np.median(dists)
                #print(f'dist_median: {dist_median}')

                filtered_polygon_points = [polygon_points[0]]
                idx = 1
                while idx < len(polygon_points):
                    idx_closest = -1
                    for idx_other in range(idx, len(polygon_points)):
                        dist = polygon_points[idx_other].distance(filtered_polygon_points[-1])
                        #print(idx, idx_other, dist)
                        if dist < dist_median * 0.3: # Need to fine-tune
                            idx_closest = idx_other
                    if idx_closest > 0:
                        idx = idx_closest + 1
                    if idx < len(polygon_points):
                        filtered_polygon_points.append(polygon_points[idx])
                    idx += 1
                polygon_points = filtered_polygon_points

                #print('polygon_points', args.dbscan_eps, args.clustering_minSamples, len(polygon_points) / 2)

                if len(polygon_points) / 2 >= 3:
                    break # found at least 4 while
                else:
                    args.clustering_minSamples += 1


            #img = plot_poly(map_image, polygon_points, show=False)
            # HoughTransform.plot(map_image, lines_filtered, show=True)
            # img = HoughTransform.plot(map_image, lines_filtered, show=False)
            #filename = file.replace('.ply', '_result.jpg')
            #cv2.imwrite(filename, img)

            vertices = image_to_vertices(map_image)

            output_dict = {
                'accessible_polygon': list(map(lambda x: x.as_list, polygon_points)),
                'vertices': vertices.tolist()
            }
    except Exception as e:
        print(e)

    return output_dict

