import enum
from abc import ABC, abstractmethod
from PyQt5 import QtCore, QtWidgets

from modules.decorators import QLongLabelDecorator, QLayoutDecorator, QRadioBtnDecorator, QShortLabelDecorator


class ViewManager:
    def __init__(self, central_widget):
        self.view: ToolsView = DefaultView()
        self.central_widget = central_widget

    def set_view(self, view):
        self.view = view

    def get_properties(self):
        return self.view.get_properties(self.central_widget)

    def get_radio_btn_state(self):
        if isinstance(self.view, EditPointsOfInterestView):
            return self.view.get_radio_btn_state()
        return None

    def get_map(self):
        return self.view.render_map()


class ToolsView(ABC):

    @abstractmethod
    def get_properties(self, central_widget):
        pass

    @abstractmethod
    def render_map(self):
        pass


class DefaultView(ToolsView):
    def __init__(self):
        self.vertical_layout = None
        self.label = None

    def get_properties(self, central_widget):
        self.vertical_layout = QLayoutDecorator(QtWidgets.QVBoxLayout(), "properties_layout")
        self.label = QShortLabelDecorator(QtWidgets.QLabel(central_widget), "default_label")

        layout = self.vertical_layout.get_element()
        label = self.label.get_element()

        _translate = QtCore.QCoreApplication.translate
        label.setText(_translate("MainWindow", "Properties - overview"))

        layout.addWidget(label)
        spacer_item = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        layout.addItem(spacer_item)
        return layout

    def render_map(self):
        pass


class EditMapView(ToolsView):
    def __init__(self):
        self.vertical_layout = None
        self.label_1 = None
        self.label_2 = None
        self.label_3 = None
        self.label_4 = None
        self.label_5 = None

    def get_properties(self, central_widget):
        self.vertical_layout = QLayoutDecorator(QtWidgets.QVBoxLayout(), "properties_layout")
        self.label_1 = QShortLabelDecorator(QtWidgets.QLabel(central_widget), "properties_label")
        self.label_2 = QLongLabelDecorator(QtWidgets.QLabel(central_widget), "info_label_1")
        self.label_3 = QLongLabelDecorator(QtWidgets.QLabel(central_widget), "info_label_2")
        self.label_4 = QLongLabelDecorator(QtWidgets.QLabel(central_widget), "info_label_3")
        self.label_5 = QLongLabelDecorator(QtWidgets.QLabel(central_widget), "info_label_4")

        layout = self.vertical_layout.get_element()
        label_1 = self.label_1.get_element()
        label_2 = self.label_2.get_element()
        label_3 = self.label_3.get_element()
        label_4 = self.label_4.get_element()
        label_5 = self.label_5.get_element()

        _translate = QtCore.QCoreApplication.translate
        label_1.setText(_translate("MainWindow", "Properties - Map editing"))
        label_2.setText(_translate("MainWindow", "Double click on line to add more points"))
        label_3.setText(_translate("MainWindow", "Click and drag corners to move them around"))
        label_4.setText(_translate("MainWindow", "Corners will snap along already existing axis"))
        label_5.setText(_translate("MainWindow", "Select a corner and press delete to remove it"))

        layout.addWidget(label_1)
        layout.addWidget(label_2)
        layout.addWidget(label_3)
        layout.addWidget(label_4)
        layout.addWidget(label_5)
        return layout

    def render_map(self):
        pass


class PointsOfInterest(enum.Enum):
    drone = 'drone'
    base_station = 'base_station'
    target = 'target'


class EditPointsOfInterestView(ToolsView):
    def __init__(self):
        self.vertical_layout = None
        self.label = None
        self.radio_btn_1 = None
        self.radio_btn_2 = None
        self.radio_btn_3 = None
        self.btn_group = None
        self.radio_btn_state = None

    def get_properties(self, central_widget):
        self.vertical_layout = QLayoutDecorator(QtWidgets.QVBoxLayout(), "properties_layout")
        self.label_1 = QShortLabelDecorator(QtWidgets.QLabel(central_widget), "properties_label")
        self.label_2 = QShortLabelDecorator(QtWidgets.QLabel(central_widget), "properties_label")
        self.radio_btn_1 = QRadioBtnDecorator(QtWidgets.QRadioButton(central_widget), "radio_btn_1")
        self.radio_btn_2 = QRadioBtnDecorator(QtWidgets.QRadioButton(central_widget), "radio_btn_2")
        self.radio_btn_3 = QRadioBtnDecorator(QtWidgets.QRadioButton(central_widget), "radio_btn_3")
        self.btn_group = QtWidgets.QButtonGroup(central_widget)

        layout = self.vertical_layout.get_element()
        label_1 = self.label_1.get_element()
        label_2 = self.label_2.get_element()
        radio_btn_1 = self.radio_btn_1.get_element()
        radio_btn_2 = self.radio_btn_2.get_element()
        radio_btn_3 = self.radio_btn_3.get_element()

        self.btn_group.addButton(radio_btn_1)
        self.btn_group.addButton(radio_btn_2)
        self.btn_group.addButton(radio_btn_3)

        _translate = QtCore.QCoreApplication.translate
        label_1.setText(_translate("MainWindow", "Add with double click. Move by drag and drop"))
        label_2.setText(_translate("MainWindow", "When selected press delete to remove it"))
        radio_btn_1.setText(_translate("MainWindow", "Target (Purple)"))
        radio_btn_2.setText(_translate("MainWindow", "Drone (Green)"))
        radio_btn_3.setText(_translate("MainWindow", "Station (Red)"))

        radio_btn_1.toggled.connect(self.set_target_state)
        radio_btn_2.toggled.connect(self.set_drone_state)
        radio_btn_3.toggled.connect(self.set_station_state)

        radio_btn_1.setChecked(True)

        layout.addWidget(label_1)
        layout.addWidget(label_2)
        layout.addWidget(radio_btn_1)
        layout.addWidget(radio_btn_2)
        layout.addWidget(radio_btn_3)
        return layout

    def set_target_state(self):
        self.radio_btn_state = PointsOfInterest.target

    def set_drone_state(self):
        self.radio_btn_state = PointsOfInterest.drone

    def set_station_state(self):
        self.radio_btn_state = PointsOfInterest.base_station

    def get_radio_btn_state(self):
        return self.radio_btn_state

    def render_map(self):
        pass
